/* This code is part of the paper
 "Space-efficient Parallel Construction of Succinct Representations
 of Suffix Tree Topologies"
 by Uwe Baier, Timo Beller and Enno Ohlebusch (to be published in JEA 2016)
 
 Copyright (c) 2016 Uwe Baier
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the
 rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 sell copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:
 
 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.
 
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 SOFTWARE.
*/
/* NOTE: this algorithm is a modified version from Fuyao Zhao and Julian Shun,
 as presented in
 "Practical Parallel Lempel-Ziv Factorization" by Julian Shun and Fuyao Zhao
 Proceedings of the IEEE Data Compression Conference (DCC), pp. 123-132, 2013.
 Code was available at https://github.com/zfy0701/Parallel-LZ77/tree/release
*/
#ifndef ANSV_PAR_HPP
#define ANSV_PAR_HPP

#include <algorithm>

#include <omp.h>

#define ANSV_LEFT(i) ((i) << 1)
#define ANSV_RIGHT(i) (((i) << 1) | 1)
#define ANSV_PARENT(i) ((i) >> 1)

template<class intT>
void ansv_par(intT *a, intT n, intT *psev, intT *nsv);

//helpers for ansv
template<class intT>
class ansv_helper {
	private:
		friend void ansv_par<intT>(intT *a, intT n, intT *psev, intT *nsv);

		static inline intT getLeft_opt(intT **table, intT depth, intT n, intT index, intT start) {
		  intT value = table[0][index];

		  intT cur = ANSV_PARENT(start), d;
		  for (d = 1; d < depth; d++) {
		    if ( (index >> d) == cur ) { //tree contains index
		      if ( (cur--) == 0 )  return n;
		    }

		    if (table[d][cur] > value) cur = ANSV_PARENT(cur);
		    else break;
		  }

		  for ( ; d > 0; d--) {
		    if (table[d - 1][ANSV_RIGHT(cur)] <= value) cur = ANSV_RIGHT(cur);
		    else cur = ANSV_LEFT(cur);
		  }
		  return cur;
		}

		static inline intT getRight_opt(intT **table, intT depth, intT n, intT index, intT start) {
		  intT value = table[0][index];
		  if (value == table[depth - 1][0]) return n;

		  intT cur = ANSV_PARENT(start), d;
		  for (d = 1; d < depth; d++) {
		    if ( (cur << d) < index ) { //current tree is left of index
		      if ( (++cur << d) >= n ) return n;
		    }

		    if (table[d][cur] >= value) cur = ANSV_PARENT(cur);
		    else break;
		  }

		  for ( ; d > 0; d--) {
		    if (table[d - 1][ANSV_LEFT(cur)] < value) cur = ANSV_LEFT(cur);
		    else cur = ANSV_RIGHT(cur);
		  }
		  return cur;
		}

		static void compute_linear(intT a[], intT n, intT psev[], intT nsv[], intT offset, intT undef) {
		  intT i, top;
		  intT *stack = new intT[n];

		  for (i = 0, top = -1; i < n; i++) {
		    while (top != -1 && a[stack[top]] > a[i]) top--;
		    if (top == -1) psev[i] = undef;
		    else psev[i] = stack[top] + offset;
		    stack[++top] = i;
		  }

		  for (i = n - 1, top = -1; i != -1; i--) {
		    while (top != -1 && a[stack[top]] >= a[i]) top--;
		    if (top == -1)  nsv[i] = undef;
		    else nsv[i] = stack[top] + offset;
		    stack[++top] = i;
		  }
		  delete[] stack;
		}
};

template<class intT>
void ansv_par(intT *a, intT n, intT *psev, intT *nsv) {
  if (n == 0) return;
  intT depth = 2;
  for (intT x = n-1; x > 1; x >>= 1, depth++);
  intT **table = new intT*[depth];

  table[0] = a;
  intT m = n;
  for (intT i = 1; i < depth; i++) {
    m = (m + 1) / 2;
    table[i] = new intT[m];
  }

  m = n;
  for (intT d = 1; d < depth; d++) {
    intT m2 = m / 2;

    #pragma omp parallel for
    for (intT i = 0; i < m2; i++) {
      table[d][i] = std::min(table[d - 1][ANSV_LEFT(i)], table[d - 1][ANSV_RIGHT(i)]);
    }

    if (m % 2) {
      table[d][m2] = table[d - 1][ANSV_LEFT(m2)];
    }

    m = (m + 1) / 2;
  }

  const intT blocksize = 8192;
  #pragma omp parallel for
  for (intT i = 0; i < n; i += blocksize) {
    intT j = i + std::min(n-i, blocksize);
    ansv_helper<intT>::compute_linear(a + i, j - i, psev + i, nsv + i, i, n);

    intT tmp = (i > 0) ? i-1 : n;
    for (intT k = i; k < j; k++) {
      if (psev[k] == n) {
        if ( tmp != n && a[tmp] > a[k]) {
          tmp = ansv_helper<intT>::getLeft_opt(table, depth, n, k, tmp);
        }
        psev[k] = tmp;
      }
    }

    tmp = j - 1;
    for (intT k = j - 1; k != i-1; k--) {
      if (nsv[k] == n) {
        if (tmp != n && a[tmp] >= a[k]) {
          tmp = ansv_helper<intT>::getRight_opt(table, depth, n, k, tmp);
        }
        nsv[k] = tmp;
      }
    }
  }

  for (intT i = 1; i < depth; i++) delete[] table[i];
  delete[] table;
}

#undef ANSV_LEFT
#undef ANSV_RIGHT
#undef ANSV_PARENT

#endif
