/* This code is part of the paper
 "Space-efficient Parallel Construction of Succinct Representations
 of Suffix Tree Topologies"
 by Uwe Baier, Timo Beller and Enno Ohlebusch (to be published in JEA 2016)
 
 Copyright (c) 2016 Uwe Baier
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the
 rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 sell copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:
 
 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.
 
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 SOFTWARE.
*/
#ifndef BPS_DFS_HPP
#define BPS_DFS_HPP

//costum declaration of a sorted stack for std namespace
namespace std {
	template<class T>
	class sorted_stack {
	public:
		typedef T value_type;
		typedef typename vector<T>::size_type size_type;
	private:
		vector<T> s; //simulate everything using a vector
	public:
		//construct a sorted multi-stack with at most n elements
		sorted_stack( size_type n ) {
			s.reserve( n );
		}

		//and add normal stack operations, see std::stack
		//TODO: add emplace and swap
		bool empty() const {
			return s.empty();
		}

		void pop() {
			s.pop_back();
		}

		void push( const value_type &val ) {
			s.push_back( val );
		}

		size_type size() const {
			return s.size();
		}

		value_type top() const {
			return s.back();
		}		
	};
};

/** construction of depth-first-search traversal bps
 * @tparam t_lcp          type of lcp array
 * @tparam value_type     type of values in lcp array. highest value must be smaller 
 *                        than max(value_type), so lcp.size() < max(value_type) must hold
 * @tparam size_type      type to index the bit vector
 * @tparam t_sorted_stack type of a sorted stack where each value is contained at most once
 * @tparam t_bv           type of bit vector
 * @param lcp   lcp array
 * @param bps   bitvector where to write balanced parentheses sequence
 * @param o_par representation of an opening parenthesis
 */
template<class t_lcp,
         class value_type = typename t_lcp::value_type,
         class size_type = typename t_lcp::size_type,
         class t_sorted_stack = std::sorted_stack<value_type>,
         class t_bv = std::vector<bool>>
void bps_dfs(const t_lcp &lcp, t_bv &bps, const bool o_par = true ) {
	const bool c_par = !o_par;

	//trim bps to maximal size of tree
	bps.resize( 4 * lcp.size() );

	if (lcp.size() == 0)	return;

	//run from back to front of lcp, enumerate intervals and count
	// opening parentheses per position i
	t_sorted_stack stack( lcp.size()+1 );
	stack.push( 0 ); //for lcp[n+1]
	size_type p = bps.size() - 1;
	for (size_type i = lcp.size() - 1; i > 0; --i) {
		//compute number of opening parentheses at position i
		size_type co = 1;  //for singleton interval
		size_type x = lcp[i]+1; //to indicate start and end of lcp-array
		while (stack.top() > x) {
			stack.pop(); ++co;
		}
		if (stack.top() < x) {
			stack.push(x);
		}
		//encode number of opening parenthesis at i as unary number
		bps[p--] = o_par;
		while (--co > 0)	bps[p--] = c_par;
	}
	//handle last value lcp[0] separate, since it virtually is a -1, but in real is a 0
	bps[p--] = o_par; //code last number of opening parenthesis
	while (stack.size() > 1) { //remove all elements except the zero from stack for next run
		stack.pop();
		bps[p--] = c_par;	//move k to first bit before unary number
	}
		

	//run from front to back of lcp, enumerate intervals,
	//write opening parentheses and leave out closing parentheses
	size_type q = 0;
	for (size_type i = 1; i < lcp.size(); ++i) {
		//compute number of opening parentheses at position i-1 using
		//the unary coding from the last step
		size_type co = 0;
		do {
			++co;
		} while (bps[++p] == c_par);

		//compute number of closing parentheses at position i-1
		size_type cc = 1; //for singleton interval
		size_type x = lcp[i]+1;
		while (stack.top() > x) {
			stack.pop();	++cc;
		}
		if (stack.top() < x) {
			stack.push(x);
		}
		//write sequence for position i-1
		while (co-- > 0)	bps[q++] = o_par;
		while (cc-- > 0)	bps[q++] = c_par;
	}
	//handle last value lcp[n+1] separate
	bps[q++] = o_par;
	while (!stack.empty()) {
		bps[q++] = c_par;
		stack.pop();
	}

	//trim bps to correct size and stop
	bps.resize(q);
}
#endif
