/* This code is part of the paper
 "Space-efficient Parallel Construction of Succinct Representations
 of Suffix Tree Topologies"
 by Uwe Baier, Timo Beller and Enno Ohlebusch (to be published in JEA 2016)
 
 Copyright (c) 2016 Uwe Baier
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the
 rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 sell copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:
 
 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.
 
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 SOFTWARE.
*/
#ifndef BPS_SADA_PAR_HEUR_HPP
#define BPS_SADA_PAR_HEUR_HPP

#include <algorithm>
#include <future>
#include <limits>
#include <memory>
#include <mutex>
#include <utility>
#include <vector>

/** heuristic construction of depth-first-search traversal bps with manual threshold
 * @tparam t_lcp type of lcp array
 * @tparam value_type type of values in lcp array. highest value must be smaller 
 *                    than max(value_type), so lcp.size() < max(value_type) must hold
 * @tparam size_type  type to index the bit vector
 * @tparam t_bv       type of bit vector
 * @tparam bv_ib_type type of words used in bit-vector, see assumption on bitvector
 * @param p         number of threads (processors) to be used for construction
 * @param threshold threshold to be used (>0, < max(value_type))
 * @param lcp       lcp array
 * @param bps       bitvector where to write balanced parentheses sequence
 * @param o_par     representation of an opening parenthesis
 */
template<class t_lcp,
         class value_type = typename t_lcp::value_type,
         class size_type = typename t_lcp::size_type,
         class t_bv = std::vector<bool>,
         class bv_ib_type = size_type>
void bps_dfs_par_heur( size_type p, size_type threshold, const t_lcp &lcp,
		t_bv &bps, const bool o_par = true );

/** heuristic construction of depth-first-search traversal bps with automatic threshold t=2*log(n)
 * @tparam t_lcp type of lcp array
 * @tparam value_type type of values in lcp array. highest value must be smaller 
 *                    than max(value_type), so lcp.size() < max(value_type) must hold
 * @tparam size_type  type to index the bit vector
 * @tparam t_bv       type of bit vector
 * @tparam bv_ib_type type of words used in bit-vector, see assumption on bitvector
 * @param p         number of threads (processors) to be used for construction
 * @param lcp       lcp array
 * @param bps       bitvector where to write balanced parentheses sequence
 * @param o_par     representation of an opening parenthesis
 */
template<class t_lcp,
         class value_type = typename t_lcp::value_type,
         class size_type = typename t_lcp::size_type,
         class t_bv = std::vector<bool>,
         class bv_ib_type = size_type>
void bps_dfs_par_heur( size_type p, const t_lcp &lcp, t_bv &bps,
		const bool o_par = true ) {
	//compute threshold based on length of input, t = 2 log(n)
	size_type threshold = 0;
	size_type n = lcp.size();
	while (n != 0) {
		n >>= 1; ++threshold;
	}
	threshold *= 2;
	bps_dfs_par_heur<t_lcp,value_type,size_type,t_bv,bv_ib_type>
		( p, threshold, lcp, bps, o_par );
}

/* Assumption on bitvector:
 * The bitvector consists of consecutive bv_ib_type words. The word of a position i
 * can be found by i / digits(bv_ib_type). Also, given such a word w, if a thread t
 * reads any n consecutive bits of w while other threads concurrently modify bits in w
 * except of those read by thread t, no side effects occur.
 */
template<class t_lcp,
         class value_type = typename t_lcp::value_type,
         class size_type = typename t_lcp::size_type,
         class t_bv = std::vector<bool>,
         class bv_ib_type = size_type>
void bps_dfs_par_heur( size_type p, size_type threshold, const t_lcp &lcp,
		t_bv &bps, const bool o_par ) {
	if (lcp.size() == 0)	return;

	p = std::max( p, (size_type)1 );
	threshold = std::max( threshold, (size_type)1 );
	size_type bs = lcp.size() / p; //blocksize
	while (bs < p) {	//ensure p * p < n
		p = (p + bs) / 2;
		bs = lcp.size() / p;
	}

	//create necessary structures
	std::vector<size_type> startpos( p + 1, 0 );
	std::vector<size_type> count( p + 1, 0 );
	std::vector<t_bv> fstack, bstack;
	fstack.reserve( p ); bstack.reserve( p );
	for (size_type i = 0; i < p; i++) {
		fstack.emplace_back( threshold+1, false );
		bstack.emplace_back( threshold+1, false );
	}

	//create stack workspace, blocksize and futures
	std::vector<value_type> stackitem( lcp.size() + threshold * p );
	std::vector<std::future<void>> t_result( p );
	bs = (size_type)(1 + lcp.size() / p); //real blocksize, ~ ceil(n / t)
	//divide lcp into initial blocks and set up initial count
	for (size_type k = 0; k < p; k++) {
		t_result[k] = std::async( std::launch::async,
		                          [&startpos, &count, &fstack, &bstack, &stackitem, &lcp]
		                          (size_type bs, size_type k, size_type t) {
			//start of parallel computation

			size_type s = k * bs;
			size_type i, e;
			value_type lcp_i; //value at index i

			if (k == fstack.size()-1) { //last block must be handled special
				i = (size_type)lcp.size();
				lcp_i = 0; //lcp[n+1] = -1
			} else { //search for exclusive right bound of block with value < threshold
				i = s + bs;
				while (lcp[i] >= t) {
					if (--i <= s)	return; //stop if bound is not found in own block
				}
				lcp_i = lcp[i]+1;
			}
			e = i;

			//run stack-based algorithm
			size_type cnt = 0; //count of parentheses in own interval
			size_type sb = i; //stackbottom
			size_type st = sb; //stacktop position (stack runs from right to left)
			stackitem[st] = lcp_i;
			for (lcp_i = lcp[--i]+1; i > s || lcp_i > t; lcp_i = lcp[--i]+1) {
				cnt += 2; //for singleton interval
				while (lcp_i < stackitem[st]) {
					++st; //pop element from stack
					++cnt; //for opening parenthesis at position i
					if (st > sb) { //if stack is empty
						stackitem[--sb] = lcp_i; //create a new stack starting one position before
						st = sb; //containing the lower value
					}
				}
				if (lcp_i > stackitem[st]) {
					stackitem[--st] = lcp_i; //push element on stack
					++cnt; //for closing parenthesis of value on stack
				}
			}
			//handle last value separate, since lcp[0] = -1 by definition
			//, but lcp[0] = 0 holds in application
			if (i == 0)	lcp_i = 0;
			cnt += 2;
			while (lcp_i < stackitem[st]) {
				++st;	++cnt;
				if (st > sb) {
					stackitem[--sb] = lcp_i;	st = sb;
				}
			}
			if (lcp_i > stackitem[st]) {
				stackitem[--st] = lcp_i;	++cnt;
			}
			
			//create forward and backward stacks from remaining values on stack
			while (st <= sb)	bstack[k][stackitem[st++]] = true;
			while (sb <= e )	fstack[k][stackitem[sb++]] = true;

			//and store remaining values
			startpos[k] = i;
			count[k] = cnt;

			//end of parallel computation
		                          }, bs, k, threshold );
	}
	//wait for threads to finish, remove empty blocks and aggregate forward stacks
	size_type _p = 1;
	t_result[0].get();
	for (size_type k = 1; k < p; k++) {
		t_result[k].get();
		if (startpos[k] != 0) {
			//copy elements to front
			startpos[_p] = startpos[k];
			count[_p] = count[k];
			fstack[_p] = fstack[k];
			bstack[_p] = bstack[k];
			//find minimal value in fstack[_p] and copy elements from fstack[_p-1]
			value_type min = 0;
			while (!fstack[_p][min]) {
				fstack[_p][min] = fstack[_p-1][min];
				++min;
			}
			//add number of closed parentheses to block _p
			while (++min <= threshold) {
				if (fstack[_p-1][min] && !bstack[_p][min])
					count[_p]++;
			}
			_p++;
		}
	}
	for (size_type k = _p-2; k != (size_type)-1; --k) { //aggregate backward stacks
		value_type min = 0;
		while (!bstack[k][min]) {
			bstack[k][min] = bstack[k+1][min];
			++min;
		}
		while (++min <= threshold) {
			if (bstack[k+1][min] && !fstack[k][min])
				count[k]++;
		}
	}

	size_type sum = 0;	//build cumulative sum of parentheses counts
	for (size_type k = 0; k <= _p; ++k) {
		size_type tmp = count[k];
		count[k] = sum;
		sum += tmp;
	}
	//now, the total number of parentheses can be fetched from count[_p]
	bps.resize( count[_p] );
	startpos[_p] = lcp.size();

	//build locks for critical bit positions
	constexpr size_type ib_bitsize = std::numeric_limits<bv_ib_type>::digits;
	typedef std::pair<std::shared_ptr<std::mutex>, size_type> ib_mutex;
	std::vector<ib_mutex> bps_ib_mtx;
	bps_ib_mtx.reserve( _p+1 );
	size_type last_ib_pos = 0;
	std::shared_ptr<std::mutex> mtxptr = std::make_shared<std::mutex>();
	for (size_type k = 0; k <= _p; ++k) {
		size_type start_ib_pos = count[k] / ib_bitsize;
		if (start_ib_pos != last_ib_pos) { 
			//bound reaches a new integer, so new mutex can be used
			mtxptr = std::make_shared<std::mutex>();
		}
		last_ib_pos = start_ib_pos;
		bps_ib_mtx.emplace_back( mtxptr, last_ib_pos );
	}

	//write sequence in parallel
	for (size_type k = 0; k < _p; ++k) {
		t_result[k] = std::async( std::launch::async,
		                          [&startpos, &count, &fstack, &bstack, &stackitem,
		                           &lcp, &bps, &bps_ib_mtx] (size_type k, size_type _p, 
			                   size_type t, const bool o_par) {
			//start of parallel computation
			const bool c_par = !o_par;

			//fixate required data structures for own run
			size_type s = startpos[k];
			size_type e = startpos[k+1];
			size_type q = count[k];
			size_type p = count[k+1];
			auto &bstack_k = bstack[(k+1) % _p]; //cyclic next bstack
			auto &fstack_k = fstack[(k+_p-1) % _p]; //cyclic previous fstack
			
			//prepare backward stack (stack runs again from left to right now)
			size_type sb = s + k*t; //leave enough space between stacks
			size_type st = sb-1;
			for (size_type i = 0; i < bstack_k.size(); i++) {
				if (bstack_k[i])	stackitem[++st] = i;
			}

			//initialize locks and lock structures
			std::unique_lock<std::mutex> bps_sm( *(bps_ib_mtx[k].first)  , std::defer_lock );
			std::unique_lock<std::mutex> bps_em( *(bps_ib_mtx[k+1].first), std::defer_lock );
			size_type bps_sm_ib = bps_ib_mtx[k].second;
			size_type bps_em_ib = bps_ib_mtx[k+1].second;

			if (bps_sm_ib == bps_em_ib)	bps_sm_ib++;
			bps_em.lock();

			//do a backward run and store number of opening parentheses
			size_type co, cc;
			value_type lcp_i;
			for (size_type i = e-1; i > s; --i) {
				co = 1;
				lcp_i = lcp[i]+1;
				while (stackitem[st] > lcp_i) {
					--st; //pop element from stack
					++co; //number of opening parentheses increases
				}
				if (stackitem[st] < lcp_i) {
					stackitem[++st] = lcp_i; //push element to stack
				}
				//write number of opening parentheses, encoded as unary number
				if ((p-co) / ib_bitsize == bps_sm_ib) { //ensure no side-effects occur in bitvector
					bps_sm.lock(); bps_sm_ib++;
				}
				bps[--p] = o_par;
				while (--co > 0)	bps[--p] = c_par;
				if (p / ib_bitsize < bps_em_ib) { //avoid side-effects in bitvector
					bps_em.unlock(); bps_em_ib = 0;
				}
			}
			//handle last value separate (lcp[0] := -1 in theory, but lcp[0] = 0 holds in practice)
			co = 1;
			lcp_i = (s == 0) ? 0 : lcp[s]+1;
			while (stackitem[st--] > lcp_i) ++co;

			//free end lock if necessary		
			if (bps_em)	bps_em.unlock();
			else		bps_em_ib = bps_ib_mtx[k+1].second;

			//prepare forward stack
			st = sb-1;
			for (size_type i = 0; i < fstack_k.size(); i++) {
				if (fstack_k[i])	stackitem[++st] = i;
			}

			//use start lock if not done yet
			if (bps_sm)			bps_sm_ib = bps_ib_mtx[k].second;
			else if (bps_sm_ib < bps_em_ib)	bps_sm.lock();

			//forward run over lcp and set up final sequence
			for (size_type i = s+1; i < e; ++i) {
				cc = 1; //for singleton interval
				lcp_i = lcp[i]+1;
				while (stackitem[st] > lcp_i) {
					--st; //pop element from stack
					++cc; //and increment number of opening parentheses at positon i-1
				}
				if (stackitem[st] < lcp_i) {
					stackitem[++st] = lcp_i; //push element to stack
				}
				//write co opening followed by cc closing parentheses at current position
				if ((q+co+cc) / ib_bitsize == bps_em_ib) {
					bps_em.lock();	bps_em_ib++;
				}

				while (co-- > 0)	bps[q++] = o_par;
				while (cc-- > 0)	bps[q++] = c_par;

				if (q / ib_bitsize > bps_sm_ib) {
					bps_sm.unlock();	bps_sm_ib = bps_em_ib;
				}
				
				//decode number of opening parenthesis at position i for next run
				co = 1;
				while (bps[p++] == c_par)	++co;
			}
			//handle last value separate, since lcp[n+1] only exists in theory
			cc = 1;
			lcp_i = (e == lcp.size()) ? 0 : lcp[e]+1;
			while (stackitem[st--] > lcp_i)	++cc;

			if (!bps_em)	bps_em.lock();
			while (co-- > 0)	bps[q++] = o_par;
			while (cc-- > 0)	bps[q++] = c_par;
			
			//remaining locks are free'd automatically

			//end of parallel computation
					}, k, _p, threshold, o_par );
	}
	//and wait for threads to finish
	for (size_type k = 0; k < _p; k++) {
		t_result[k].get();
	}
}

#endif
