/* This code is part of the paper
 "Space-efficient Parallel Construction of Succinct Representations
 of Suffix Tree Topologies"
 by Uwe Baier, Timo Beller and Enno Ohlebusch (to be published in JEA 2016)
 
 Copyright (c) 2016 Uwe Baier
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the
 rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 sell copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:
 
 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.
 
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 SOFTWARE.
*/
#ifndef BPS_SADA_PAR_OPT_HPP
#define BPS_SADA_PAR_OPT_HPP

#include <algorithm>
#include <iterator>
#include <limits>
#include <memory>
#include <vector>

#include <omp.h>

#include "ansv_par.hpp"

/** theoretically optimal construction of depth-first-search traversal bps
 * @tparam value_type type of values in lcp array. highest value must be smaller 
 *                    than max(value_type), so lcp.size() < max(value_type) must hold
 * @tparam size_type  type to index the bit vector
 * @tparam t_bv       type of bit vector
 * @tparam bv_ib_type type of words used in bit-vector, see assumption on bitvector
 * @param p         number of threads (processors) to be used for construction
 * @param lcp       lcp array
 * @param bps       bitvector where to write balanced parentheses sequence
 * @param o_par     representation of an opening parenthesis
 */
template<class value_type,
         class size_type = typename std::vector<value_type>::size_type,
         class t_bv = std::vector<bool>,
         class bv_ib_type = size_type>
/* Assumption on bitvector:
 * The bitvector consists of consecutive bv_ib_type words. The word of a position i
 * can be found by i / digits(bv_ib_type). Also, given such a word w, if a thread t
 * reads any n consecutive bits of w while other threads concurrently modify bits in w
 * except of those read by thread t, no side effects occur.
 */
void bps_dfs_par_opt( size_type p, std::vector<value_type> &lcp,
		t_bv &bps, const bool o_par = true ) {
	const bool c_par = !o_par;

	size_type bs = lcp.size() / p; //blocksize
	while (bs < p) {	//ensure p * p < n
		p = (p + bs) / 2;
		bs = lcp.size() / p;
	}

	//do some omp setup
	omp_set_dynamic( 0 );
	omp_set_num_threads( p );
	
	if (lcp.size() == 0)	return;

	//create necessary vectors
	std::vector<size_type> Co;
	std::vector<size_type> Cc;

	{ //count number of opening / closing parentheses per position in parallel
		std::vector<value_type> psev( lcp.size() );
		std::vector<value_type> nsv( lcp.size() );
		ansv_par( lcp.data(), (value_type)lcp.size(), psev.data(), nsv.data() );

		Co.resize( lcp.size(), (size_type)1 );
		Cc.resize( lcp.size(), (size_type)1 );

		#pragma omp parallel for
		for (size_type i = 1; i < lcp.size(); i++) {
			if (lcp[i] != lcp[psev[i]] || psev[i] == 0) {
				size_type io = psev[i];
				size_type ic = nsv[i] - 1;
				#pragma omp atomic
				++Co[io];
				#pragma omp atomic
				++Cc[ic];
			}
		}
	}

	std::vector<size_type> &PS = Cc;
	{ //build prefix sum of overall parentheses, and store it to Cc
		std::vector<size_type> count( p );
		bs = (size_type)(1 + lcp.size() / p); //real blocksize, ~ ceil(n / t)
		#pragma omp parallel for
		for (size_type t = 0; t < p; t++) { //count in each block
			size_type e = (t == p-1) ? lcp.size() : (t+1) * bs;
			size_type cnt = 0;
			for (size_type i = t * bs; i < e; ++i) {
				cnt += Co[i] + Cc[i];
			}
			count[t] = cnt;
		}

		size_type sum = 0; //build prefix sum of blocks
		for (size_type t = 0; t < p; t++) {
			size_type tmp = count[t];
			count[t] = sum;
			sum += tmp;
		}

		#pragma omp parallel for
		for (size_type t = 0; t < p; t++) {
			size_type e = (t == p-1) ? lcp.size() : (t+1) * bs;
			size_type cnt = count[t];
			for (size_type i = t * bs; i < e; ++i) {
				cnt += Co[i] + Cc[i];
				PS[i] = cnt;
			}
		}
		bps.resize( sum ); //trim bitvector to correct size
	}

	//find almost optimal start positions for each thread
	std::vector<value_type> startpos( p+1 );
	#pragma omp parallel for
	for (size_type t = 0; t < p; t++) { //use binary search for each thread
		size_type ds = t * (bps.size() / p); //desired start position
		auto it = std::upper_bound( PS.begin(), PS.end(), ds );
		startpos[t] = (value_type) std::distance( Cc.begin(), it );
	}
	startpos[p] = lcp.size();

	//create locks for critical positions in bitvector
	constexpr size_type ib_bitsize = std::numeric_limits<bv_ib_type>::digits;
	typedef std::pair<size_type, std::shared_ptr<omp_lock_t>> bps_lock;
	std::vector<bps_lock> locks;
	locks.reserve( p+1 );
	auto lockdeleter = []( omp_lock_t *l ) {
		omp_destroy_lock(l);
		delete l;
	};
	std::shared_ptr<omp_lock_t> omp_lock( new omp_lock_t, lockdeleter );
	omp_init_lock( omp_lock.get() );
	locks.push_back( std::make_pair( (size_type)0, omp_lock ) );
	size_type last_ib_pos = 0;
	for (size_type t = 1; t <= p; t++) {
		size_type sp = (startpos[t] > 0) ? startpos[t]-1 : 0;
		size_type ib_pos = PS[sp] / ib_bitsize;
		if (ib_pos != last_ib_pos) { //bits in different blocks, use a new lock
			omp_lock = std::shared_ptr<omp_lock_t>( new omp_lock_t, lockdeleter );
			omp_init_lock( omp_lock.get() );
		}
		last_ib_pos = ib_pos;
		locks.push_back( std::make_pair( ib_pos, omp_lock ) );
	}

	//write sequence using Co and PS
	#pragma omp parallel for
	for (size_type t = 0; t < p; t++) {
		size_type e = startpos[t+1];
		size_type q = (startpos[t] > 0) ? PS[startpos[t]-1] : 0;

		//care for locks
		size_type bps_sm_ib = locks[t].first;
		size_type bps_em_ib = locks[t+1].first;
		auto bps_sm = locks[t].second.get();
		auto bps_em = locks[t+1].second.get();
		if (bps_sm_ib == bps_em_ib)	bps_sm_ib = lcp.size();
		else				omp_set_lock( bps_sm );
		
		for (size_type i = startpos[t]; i < e; i++) {
			if (PS[i] / ib_bitsize == bps_em_ib) { //avoid side effects
				omp_set_lock( bps_em );
				bps_em_ib++;
			}

			//write sequence
			for (size_type j = 0; j < Co[i]; j++) { //write Co[i] opening parentheses
				bps[q++] = o_par;
			}
			do {	//and write closing parentheses till end
				bps[q++] = c_par;
			} while (q < PS[i]);

			if (PS[i] / ib_bitsize > bps_sm_ib) { //avoid side effects
				omp_unset_lock( bps_sm );
				bps_sm_ib = lcp.size();
			}
		}

		if (bps_em_ib != locks[t+1].first) { //thread owns end lock
			omp_unset_lock( bps_em );
		}
	}
}

#endif
