/* This code is part of the paper
 "Space-efficient Parallel Construction of Succinct Representations
 of Suffix Tree Topologies"
 by Uwe Baier, Timo Beller and Enno Ohlebusch (to be published in JEA 2016)
 
 Copyright (c) 2016 Uwe Baier
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the
 rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 sell copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:
 
 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.
 
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 SOFTWARE.
*/
#ifndef BPS_SCT_HPP
#define BPS_SCT_HPP

#include <stack>
#include <vector>

//costum declaration of a sorted multistack for std namespace
namespace std {
	template<class T>
	class sorted_multi_stack {
	public:
		typedef T value_type;
		typedef typename vector<T>::size_type size_type;
	private:
		vector<T> s; //simulate everything using a vector
	public:
		//construct a sorted multi-stack with at most n elements
		sorted_multi_stack( size_type n ) {
			s.reserve( n );
		}

		//and add normal stack operations, see std::stack
		//TODO: add emplace and swap
		bool empty() const {
			return s.empty();
		}

		void pop() {
			s.pop_back();
		}

		void push( const value_type &val ) {
			s.push_back( val );
		}

		size_type size() const {
			return s.size();
		}

		value_type top() const {
			return s.back();
		}		
	};
};

/** construction of super cartesian tree bps
 * @tparam t_lcp                type of lcp array
 * @tparam value_type           type of values in lcp array. highest value must be smaller 
 *                              than max(value_type), so lcp.size() < max(value_type) must hold
 * @tparam size_type            type to index the bit vectors
 * @tparam t_sorted_multi_stack type of a sorted stack where a value can be contained multiple times
 * @tparam t_bv                 type of bit vector
 * @param lcp   lcp array
 * @param bpr   bitvector where to write balanced parentheses sequence
 * @param b     bitvector for enhancing sequence
 * @param o_par representation of an opening parenthesis
 */
template<class t_lcp,
         class value_type = typename t_lcp::value_type,
         class size_type = typename t_lcp::size_type,
         class t_sorted_multi_stack = std::sorted_multi_stack<value_type>,
         class t_bv = std::vector<bool>>
void bps_sct(t_lcp &lcp, t_bv &bpr, t_bv &b, const bool o_par = true ) {
	const bool c_par = !o_par;

	//trim bpr and b to correct size
	bpr.resize( 2 * lcp.size() );
	b.resize( lcp.size() );

	if (lcp.size() == 0)	return;

	//set up a stack and run algorithm
	t_sorted_multi_stack stack( lcp.size() );
	//run initial step
	stack.push( lcp[0] );
	size_t j = 0, k = 0;
	bpr[j++] = o_par;
	//further steps
	for (size_type i = 1; i < lcp.size(); ++i) {
		value_type x = lcp[i];
		while (x < stack.top()) {
			value_type y = stack.top();
			stack.pop();

			bpr[j++] = c_par;
			b[k++] = (bool)(y != stack.top());
		}
		stack.push( x );
		bpr[j++] = o_par;
	}
	//and remaining steps that are altogether placed on the stack
	do {
		size_type y = stack.top();
		stack.pop();
		
		bpr[j++] = c_par;
		b[k++]     = (bool)(stack.empty() || y != stack.top());
	} while (!stack.empty());
}
#endif
