/* This code is part of the paper
 "Space-efficient Parallel Construction of Succinct Representations
 of Suffix Tree Topologies"
 by Uwe Baier, Timo Beller and Enno Ohlebusch (to be published in JEA 2016)
 
 Copyright (c) 2016 Uwe Baier
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the
 rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 sell copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:
 
 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.
 
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 SOFTWARE.
*/
#ifndef BPS_SCT_PAR_HEUR_HPP
#define BPS_SCT_PAR_HEUR_HPP

#include <algorithm>
#include <future>
#include <limits>
#include <list>
#include <memory>
#include <mutex>
#include <stack>
#include <utility>
#include <vector>

/** parallel heuristical construction of super cartesian tree bps with manual threshold
 * @tparam t_lcp type of lcp array
 * @tparam value_type type of values in lcp array. highest value must be smaller 
 *                    than max(value_type), so lcp.size() < max(value_type) must hold
 * @tparam size_type  type to index the bit vectors
 * @tparam t_bv       type of bit vector
 * @tparam bv_ib_type type of words used in bit-vector, see assumption on bitvector
 * @param p         number of threads (processors) to be used for construction
 * @param threshold threshold to be used (>0, < max(value_type))
 * @param lcp       lcp array
 * @param bpr       bitvector where to write balanced parentheses sequence
 * @param b         bitvector for enhancing sequence
 * @param o_par     representation of an opening parenthesis
 */
template<class t_lcp,
         class value_type = typename t_lcp::value_type,
         class size_type = typename t_lcp::size_type,
         class t_bv = std::vector<bool>,
         class bv_ib_type = size_type>
void bps_sct_par_heur(size_type p, size_type threshold, const t_lcp &lcp, 
		t_bv &bpr, t_bv &b, const bool o_par = true );

/** parallel heuristical construction of super cartesian tree bps with automatic threshold t=2*log(n)
 * @tparam t_lcp type of lcp array
 * @tparam value_type type of values in lcp array. highest value must be smaller 
 *                    than max(value_type), so lcp.size() < max(value_type) must hold
 * @tparam size_type  type to index the bit vectors
 * @tparam t_bv       type of bit vector
 * @tparam bv_ib_type type of words used in bit-vector, see assumption on bitvector
 * @param p         number of threads (processors) to be used for construction
 * @param threshold threshold to be used (>0, < max(value_type))
 * @param lcp       lcp array
 * @param bpr       bitvector where to write balanced parentheses sequence
 * @param b         bitvector for enhancing sequence
 * @param o_par     representation of an opening parenthesis
 */
template<class t_lcp,
         class value_type = typename t_lcp::value_type,
         class size_type = typename t_lcp::size_type,
         class t_bv = std::vector<bool>,
         class bv_ib_type = size_type>
void bps_sct_par_heur(size_type p, const t_lcp &lcp, t_bv &bpr, t_bv &b, 
		const bool o_par = true ) {
	//compute threshold based on length of input, t = 2 log(n)
	size_type threshold = 0;
	size_type n = lcp.size();
	while (n != 0) {
		n >>= 1; ++threshold;
	}
	threshold *= 2;
	bps_sct_par_heur<t_lcp,value_type,size_type,t_bv,bv_ib_type>
		( p, threshold, lcp, bpr, b, o_par );
}

/* Assumption on bitvector:
 * The bitvector consists of consecutive bv_ib_type words. The word of a position i
 * can be found by i / digits(bv_ib_type). Also, given such a word w, if a thread t
 * reads any n consecutive bits of w while other threads concurrently modify bits in w
 * except of those read by thread t, no side effects occur.
 */
template<class t_lcp,
         class value_type = typename t_lcp::value_type,
         class size_type = typename t_lcp::size_type,
         class t_bv = std::vector<bool>,
         class bv_ib_type = size_type>
void bps_sct_par_heur(size_type p, size_type threshold, const t_lcp &lcp, 
		t_bv &bpr, t_bv &b, const bool o_par ) {
	if (lcp.size() == 0)	return;

	p = std::max( p, (size_type)1 );
	threshold = std::max( threshold, (size_type)1 );
	size_type bs = lcp.size() / p; //blocksize
	while (bs < p) {	//ensure p * p < n
		p = (p + bs) / 2;
		bs = lcp.size() / p;
	}

	//trim bpr and b to correct size
	if (bpr.size() != 2 * (size_type)lcp.size())
		bpr.resize( 2 * (size_type)lcp.size() );
	if (b.size() != lcp.size())
		b.resize( lcp.size() );

	//create necessary structures
	std::vector<size_type> startpos( p + 1, lcp.size() );
	std::vector<size_type> dist( p + 1, 0 );

	//create stack workspace, blocksize and futures
	std::vector<value_type> stackitem( (size_type)lcp.size() + 2 * threshold * p );
	std::vector<std::future<void>> t_result( p );
	bs = (size_type)(1 + lcp.size() / p); //real blocksize, ~ ceil(n / t)

	//compute forward stacks and detect minimum of block
	for (size_type k = 0; k < p; k++) {
		t_result[k] = std::async( std::launch::async,
		                          [&lcp, &stackitem, &startpos]
		                          ( size_type bs, size_type k, size_type t ) {
			//start of parallel computation
			size_type s = k * bs;
			size_type e = std::min( s+bs, (size_type)lcp.size() );
			size_type st = k * t; //beginning position for stats in own block
			//prepare own stack block (not necessary as vector initializes everything with zero)
			//for (size_type i = 0; i < t; i++) {
			//	stackitem[st+i] = 0;
			//}
			//and fill stack block
			value_type min = t-1;
			while (e-- != s) {
				if (lcp[e] <= min) {
					min = lcp[e];
					stackitem[st+min]++;
					startpos[k] = e;
				}
			}				
			//end of parallel computation
		                          }, bs, k, threshold );
	}

	//wait for threads to finish and aggregate stacks
	size_type _p = 1;
	size_type l = 0; //last non-empty block
	t_result[0].get();
	for (size_type k = 1; k < p; k++) {
		t_result[k].get();
		if (startpos[k] != lcp.size()) { //block non-empty
			//aggregate stack with stack from previous block
			size_type l_sb = l * threshold;
			size_type k_sb = k * threshold;
			for (size_type i = 0; i <= lcp[startpos[k]]; i++) {
				stackitem[k_sb+i] += stackitem[l_sb+i];
			}
			l = k;
			//move startpos to front
			startpos[_p++] = startpos[k];
		}
	}
	startpos[_p] = lcp.size();

	//compute distances and move stacks to their final destination,
	// stored as (count,value) - pairs
	size_type k = _p-1;
	while (l-- != 0) { //ignore last non-empty block
		size_type l_sb = l * threshold;
		if (stackitem[l_sb] != 0) { //block is non-empty
			//move stack to start of own block, include padding
			size_type st = startpos[k] + 2 * k * threshold - 1;
			for (value_type i = 0; i <= lcp[startpos[k]]; i++) {
				if (stackitem[l_sb+i] > 0) { //value i is contained in block
					//compute distance
					dist[k] += stackitem[l_sb+i];

					//store (count,value) - pair in start of own stack environment
					stackitem[++st] = stackitem[l_sb+i];
					stackitem[++st] = i;
				}
			}
			//store start address of local stack on first non-empty position
			stackitem[startpos[k] + 2 * (k+1) * threshold] = st;
			--k; //and write next distance in previous block
		}
	}
	stackitem[2*threshold] = (value_type)-1; //first block has no stack at start

	//build locks for critical bit positions
	constexpr size_type bv_ib_bitsize = std::numeric_limits<bv_ib_type>::digits;
	typedef std::pair<std::shared_ptr<std::mutex>, size_type> ib_mutex;

	//mutexes for bpr
	std::vector<ib_mutex> bpr_ib_mtx;
	bpr_ib_mtx.reserve( _p+1 );
	size_type last_ib_pos = 0;
	std::shared_ptr<std::mutex> mtxptr = std::make_shared<std::mutex>();
	for (size_type k = 0; k <= _p; k++) {
		size_type start_ib_pos = (2 * startpos[k] - dist[k]) / bv_ib_bitsize;
		if (start_ib_pos != last_ib_pos) { 
			//bound reaches a new integer, so new mutex can be used
			mtxptr = std::make_shared<std::mutex>();
		}
		last_ib_pos = start_ib_pos;
		bpr_ib_mtx.push_back( ib_mutex( mtxptr, last_ib_pos ) );
	}

	//mutexes for b
	std::vector<ib_mutex> b_ib_mtx;
	b_ib_mtx.reserve( _p+1 );
	last_ib_pos = 0;
	mtxptr = std::make_shared<std::mutex>();
	for (size_type k = 0; k <= _p; k++) {
		size_type start_ib_pos = (startpos[k] - dist[k]) / bv_ib_bitsize;
		if (start_ib_pos != last_ib_pos) {
			//bound reaches a new integer, so new mutex can be used
			mtxptr = std::make_shared<std::mutex>();
		}
		last_ib_pos = start_ib_pos;
		b_ib_mtx.push_back( ib_mutex( mtxptr, last_ib_pos ) );
	}

	//create bitsequence in parallel
	for (size_type k = 0; k < _p; k++) {
		t_result[k] = std::async( std::launch::async,
		                          [&lcp, &stackitem, &bpr, &b, &bpr_ib_mtx, &b_ib_mtx]
		                          ( size_type k, size_type t, size_type s, size_type e,
		                            size_type ds, size_type de, const bool o_par ) {
			//start of parallel code
			const bool c_par = !o_par;

			//compute start positions in bpr and b
			size_type bpr_pos = 2 * s - ds;
			size_type b_pos = s - ds;
			size_type bpr_endpos = 2 * e - de;
			size_type b_endpos = e - de;

			//get stack top position
			size_type st = stackitem[s + 2 * (k+1) * t];
			size_type cst = st; //top of "counting" stack

			//initialize locks and lock structures
			std::unique_lock<std::mutex> bpr_sm( *(bpr_ib_mtx[k].first)  , std::defer_lock );
			std::unique_lock<std::mutex> bpr_em( *(bpr_ib_mtx[k+1].first), std::defer_lock );
			size_type bpr_sm_ib = bpr_ib_mtx[k].second;
			size_type bpr_em_ib = bpr_ib_mtx[k+1].second;
			if (bpr_sm_ib == bpr_em_ib)	bpr_em_ib++;

			std::unique_lock<std::mutex> b_sm( *(b_ib_mtx[k].first)  , std::defer_lock );
			std::unique_lock<std::mutex> b_em( *(b_ib_mtx[k+1].first), std::defer_lock );
			size_type b_sm_ib = b_ib_mtx[k].second;
			size_type b_em_ib = b_ib_mtx[k+1].second;
			if (b_sm_ib == b_em_ib)	b_em_ib++;

			//lock both start mutexes simultaneous, to avoid deadlocks with other threads
			std::lock( bpr_sm, b_sm );

			//run stack-based algorithm
			stackitem[++st] = lcp[s++];
			while (s < e) {
				size_type ps2w = 1; //parentheses to write this iteration
				while (stackitem[st] > lcp[s]) {
					size_type bs2w = 0; //bits in vector b to write
					value_type lcptop = stackitem[st];
					do {	//pop elements until new value
						if (st == cst) { //reached counting stack
							bs2w += stackitem[cst-1];
							cst -= 2;	st -= 2;
						} else {
							++bs2w;	--st;
						}
					} while (stackitem[st] == lcptop);

					ps2w += bs2w; //add number of closing parentheses

					//write b bitsequence
					if ((b_pos+bs2w) / bv_ib_bitsize == b_em_ib) {
						b_em.lock();	b_em_ib++;
					}
					while (--bs2w > 0)	b[b_pos++] = false;
					b[b_pos++] = true;
					if (b_pos / bv_ib_bitsize > b_sm_ib) {
						b_sm.unlock();	b_sm_ib = b_em_ib;
					}
				}
				
				//write parentheses
				if ((bpr_pos+ps2w) / bv_ib_bitsize == bpr_em_ib) {
					bpr_em.lock();	bpr_em_ib++;
				}
				bpr[bpr_pos++] = o_par;
				while (--ps2w > 0)	bpr[bpr_pos++] = c_par;
				if (bpr_pos / bv_ib_bitsize > bpr_sm_ib) {
					bpr_sm.unlock();	bpr_sm_ib = bpr_em_ib;
				}

				stackitem[++st] = lcp[s++]; //push value to stack
			}
			//handle last value "by hand", as lcp[n+1] only exists in theory

			//write parentheses in bpr
			if (bpr_pos / bv_ib_bitsize != bpr_ib_mtx[k+1].second) {
				bpr_em.lock();
			}
			bpr[bpr_pos++] = o_par;
			while (bpr_pos < bpr_endpos)	bpr[bpr_pos++] = c_par;

			if (bpr_sm)	bpr_sm.unlock(); //free locks
			if (bpr_em)	bpr_em.unlock();

			//write b bitvector
			if (b_pos / bv_ib_bitsize != b_ib_mtx[k+1].second) {
				b_em.lock();
			}
			while (st != cst) { //write from normal stack
				if (b_pos == b_endpos)	break;
				b[b_pos++] = !(stackitem[st] == stackitem[st-1]);
				--st;
			}
			while (b_pos < b_endpos) { //write from counting stack
				size_type bs2w = stackitem[st-1];
				while (--bs2w > 0)	b[b_pos++] = false;
				b[b_pos++] = true;
				st -= 2;
			}

			//remaining locks are free'd automatically

			//end of parallel code
		                          }, k, threshold, startpos[k], startpos[k+1], 
		                             dist[k], dist[k+1], o_par);
	}

	//and wait for threads to finish
	for (size_type k = 0; k < _p; k++) {
		t_result[k].get();
	}
}

#endif
