/* This code is part of the paper
 "Space-efficient Parallel Construction of Succinct Representations
 of Suffix Tree Topologies"
 by Uwe Baier, Timo Beller and Enno Ohlebusch (to be published in JEA 2016)
 
 Copyright (c) 2016 Uwe Baier
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the
 rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 sell copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:
 
 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.
 
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 SOFTWARE.
*/
#ifndef BPS_SCT_PAR_OPT_HPP
#define BPS_SCT_PAR_OPT_HPP

#include <stdint.h>
#include <memory>
#include <utility>
#include <vector>

#include <omp.h>

#include <sdsl/int_vector.hpp>

#include "ansv_par.hpp"
#include "dist_par.hpp"

/** parallel optimal construction of super cartesian tree bps
 * @tparam value_type           type of values in lcp array. highest value must be smaller 
 *                              than max(value_type), so lcp.size() < max(value_type) must hold
 * @param lcp   lcp array
 * @param bpr   bitvector where to write balanced parentheses sequence
 * @param b     bitvector for enhancing sequence
 * opening parentheses are represented with a '1', closing parentheses with a '0'
 */
template<class value_type>
void bps_sct_par_opt( uint64_t p, std::vector<value_type> &lcp, sdsl::bit_vector &bpr, 
                    sdsl::bit_vector &b ) {
	//do some omp setup
	omp_set_dynamic( 0 );
	omp_set_num_threads( p );

	//trim bpr and b to correct size
	if (bpr.size() != 2 * (uint64_t)lcp.size())
		bpr.resize( 2 * (uint64_t)lcp.size() );
	if (b.size() != lcp.size())
		b.resize( lcp.size() );

	//compute previous smaller or equal values using shun's algorithm
	std::vector<value_type> psev( lcp.size() );
	std::vector<value_type> nsv( lcp.size() );
	ansv_par( lcp.data(), (value_type)lcp.size(), psev.data(), nsv.data() );

	//compute distances of psev
	std::vector<value_type> dist( lcp.size() + 1 );
	dist_par( psev.data(), (value_type)lcp.size(), dist.data() );
	//note that dist[n] = 0 must hold

	//initialize bpr with closing brackets only
	uint64_t num_blocks = bpr.size() >> 6; //number of 64 bit blocks
	#pragma omp parallel for
	for (uint64_t i = 0; i < num_blocks; ++i) {
		bpr.set_int( i << 6, (uint64_t)0, 64 );
	}
	bpr.set_int( num_blocks << 6, (uint64_t)0, bpr.size() - (num_blocks << 6) );


	//create locks for critical positions when writing opening parenthesis of bpr
	typedef std::pair<uint64_t, std::shared_ptr<omp_lock_t>> bpr_lock;
	std::vector<bpr_lock> locks;
	locks.reserve( p + 1 );
	auto lockdeleter = []( omp_lock_t *l ) {
		omp_destroy_lock(l);
		delete l;
	};
	
	uint64_t blocksize = lcp.size() / p;
	uint64_t remainder = lcp.size() % p;

	std::shared_ptr<omp_lock_t> omp_lock( new omp_lock_t, lockdeleter );
	omp_init_lock( omp_lock.get() );

	uint64_t bpr_last_pos = 0;
	uint64_t lcpstart = 0;

	for (uint64_t i = 0; i <= p; i++) {
		uint64_t bpr_pos = 2 * (uint64_t)lcpstart - dist[lcpstart];
		if ( (bpr_pos >> 6) != (bpr_last_pos >> 6) ) { //bits in different blocks,
			//so a new lock can be used
			omp_lock = std::shared_ptr<omp_lock_t>( new omp_lock_t, lockdeleter );
			omp_init_lock( omp_lock.get() );
		}
		locks.push_back( std::make_pair( lcpstart, omp_lock ) );
		bpr_last_pos = bpr_pos;

		lcpstart += blocksize;
		if (i < remainder)	lcpstart++;
	}

	//fill bpr with the missing opening parenthesis
	#pragma omp parallel for
	for (uint64_t t = 0; t < p; t++) {
		uint64_t lcpstart = locks[t].first;
		uint64_t lcpend = locks[t+1].first;
		
		//blocks of start and end lock
		uint64_t bpr_sl_block = (2 * lcpstart - dist[lcpstart]) >> 6;
		uint64_t bpr_el_block = (2 * lcpend - dist[lcpend]) >> 6;
		
		//set start lock as long as necessary
		uint64_t bpr_pos = 2 * lcpstart - dist[lcpstart];
		omp_set_lock( locks[t].second.get() );
		while (lcpstart < lcpend) {
			if ( (bpr_pos >> 6) != bpr_sl_block )
				break;
			bpr[bpr_pos] = true; //set opening parenthesis
			++lcpstart;
			bpr_pos = 2 * lcpstart - dist[lcpstart];
		}
		omp_unset_lock( locks[t].second.get() );

		//process until end lock needs to be set
		while (lcpstart < lcpend) {
			if ( (bpr_pos >> 6) == bpr_el_block )
				break;
			bpr[bpr_pos] = true;	
			++lcpstart;
			bpr_pos = 2 * lcpstart - dist[lcpstart];
		}

		//set end lock and process the remaining parenthesis
		omp_set_lock( locks[t+1].second.get() );
		while (lcpstart < lcpend) {
			bpr[bpr_pos] = true;	
			++lcpstart;
			bpr_pos = 2 * lcpstart - dist[lcpstart];
		}
		omp_unset_lock( locks[t+1].second.get() );
	}

	//use psev to store if lcp[i] == lcp[psev[i]]
	psev[0] = 2;
	#pragma omp parallel for
	for (uint64_t i = 1; i < lcp.size(); i++) {
		psev[i] = (lcp[i] != lcp[psev[i]]) ? 2 : 0;
	}

	//put the information to the correct positions into psev
	//using nsv, dist and atomic incrementation
	#pragma omp parallel for
	for (uint64_t i = 0; i < lcp.size(); i++) {
		uint64_t b_pos = nsv[i] - 1 - dist[i];
		if ( (psev[i] & 2) != 0)	{
			#pragma omp atomic
			psev[b_pos]++;
		}
	}

	//finally, impact psev into the bit vector b
	num_blocks = lcp.size() >> 6;
	#pragma omp parallel for
	for (uint64_t i = 0; i < num_blocks; i++) {
		uint64_t b_pos = i << 6;
		for (uint64_t j = 0; j < 64; j++) {
			b[b_pos+j] = (bool)( (psev[b_pos+j] & 1) == 1 );
		}
	}
	for (uint64_t b_pos = lcp.size() & sdsl::bits::lo_unset[6]; 
	     b_pos < lcp.size(); b_pos++) {
		b[b_pos] = (bool)( (psev[b_pos] & 1) == 1 );
	}
}

#endif
