/* This code is part of the paper
 "Space-efficient Parallel Construction of Succinct Representations
 of Suffix Tree Topologies"
 by Uwe Baier, Timo Beller and Enno Ohlebusch (to be published in JEA 2016)
 
 Copyright (c) 2016 Uwe Baier
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the
 rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 sell copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:
 
 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.
 
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 SOFTWARE.
*/
#ifndef DIST_PAR_HPP
#define DIST_PAR_HPP

#include <algorithm>
#include <vector>

#include <omp.h>

template<class size_type>
void dist_par( const size_type *psev, size_type n, size_type *dist ) {
	if (n == 0)	return;
		
	size_type blocksize = n / omp_get_max_threads();
	if (n % omp_get_max_threads() > 0)	blocksize++;
	size_type useable_threads = n / blocksize;
	if (n % blocksize > 0)			useable_threads++;

	//step 1: get index of the last prev - pointer what jumps out
	// of own block, and store it to min array
	std::vector<size_type> min( useable_threads );
	#pragma omp parallel for
	for (size_type t = 0; t < useable_threads; t++) {

		size_type b_start = t * blocksize;
		size_type b_end = b_start + std::min( blocksize, n - b_start );

		//find last psev value what falls out of block
		size_type i = b_end;
		do {
			i--;
		} while (psev[i] / blocksize == t);
		min[t] = i;

		//and compute local distances starting from that point
		dist[i] = 0;
		while (++i < b_end)	{
			dist[i] = dist[psev[i]] + 1;
		}
	}

	//step 2: compute distances from minimum to minimum of each block
	//by use of local distances
	//this block can be done sequential, since the time to start 
	//threads again would be higher than simply computing directly
	for (size_type t = 1; t < useable_threads; t++) {
		size_type prev_block = psev[min[t]] / blocksize;
		dist[min[t]] = 1                       //jump to previous block
		             + dist[min[prev_block]];  //jump to start

		//add a local jump to minimum in block if necessary
		if (min[prev_block] != psev[min[t]]) {
			dist[min[t]] += dist[psev[min[t]]];     //jump to minimum in block
		}
	}

	//step 3: make all local distances to global distances
	// (first block already has global distances)
	#pragma omp parallel for
	for (size_type t = 1; t < useable_threads; t++) {
		size_type b_end = t * blocksize + std::min( blocksize, n - t * blocksize );
		size_type mindist = dist[min[t]];

		for (size_type i = min[t]+1; i < b_end; i++) {
			dist[i] += mindist;
		}
	}

	//step 4: compute the remaining distances
	#pragma omp parallel for
	for (size_type t = 1; t < useable_threads; t++) {
		for (size_type i = t * blocksize; i < min[t]; i++) {
			dist[i] = dist[psev[i]] + 1;
		}
	}
}	

#endif
