/* This code is part of the paper
 "Space-efficient Parallel Construction of Succinct Representations
 of Suffix Tree Topologies"
 by Uwe Baier, Timo Beller and Enno Ohlebusch (to be published in JEA 2016)
 
 Copyright (c) 2016 Uwe Baier
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the
 rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 sell copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:
 
 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.
 
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 SOFTWARE.
*/
/*
 NOTE: this algorithm is part of Julian Shun's parallel suffix tree algorithm,
 as presented in
 "A Simple Parallel Cartesian Tree Algorithm and its Application to Parallel
 Suffix Tree Construction" by Julian Shun and Guy Blelloch
 ACM Transactions on Parallel Computing (TOPC), Vol. 1 Issue 1, Article No. 8, 2014.
 Code was available at http://www.eecs.berkeley.edu/~jshun/suffixTree.tar.gz
*/
#ifndef SHUN_H
#define SHUN_H

#include <stdint.h>
#include <malloc.h>
#include <cilk/cilk.h>

namespace shun {

typedef int64_t intT;
typedef uint64_t uintT;

struct node { uintT firstChar; uintT parent; uintT value; };


inline long getRoot(node* nodes, long i) {
  long root = nodes[i].parent;
  while (root != 0 && nodes[nodes[root].parent].value == nodes[root].value)
    root = nodes[root].parent;
  return root;
}

void merge(node* N, uintT left, uintT right) {
  uintT head;
  if (N[left].value > N[right].value) {
    head = left; left = N[left].parent;}
  else {head = right; right= N[right].parent;}

  while(1) {
    if (left == 0) {N[head].parent = right; break;}
    if (right == 0) {N[head].parent = left; break;}
    if (N[left].value > N[right].value) {
      N[head].parent = left; left = N[left].parent;}
    else {N[head].parent = right; right = N[right].parent;}
    head = N[head].parent;}
}

void cartesianTree(node* Nodes, uintT s, uintT n) {
  if (n < 2) return;
  if(n == 2) {
    if (Nodes[s].value > Nodes[s+1].value) Nodes[s].parent=s+1;
    else Nodes[s+1].parent=s;
    return;
  }
  if (n > 1000){
    cilk_spawn cartesianTree(Nodes,s,n/2);
    cartesianTree(Nodes,s+n/2,n-n/2);
    cilk_sync;
  } else {
    cartesianTree(Nodes,s,n/2);
    cartesianTree(Nodes,s+n/2,n-n/2);
  }
  merge(Nodes,s+n/2-1,s+n/2);
}

void prepare(uintT* SA, uintT* LCP, long n, node *nodes){
  //long m = 2*n;

  //initialize nodes
  cilk_for(long i=1; i<n; i++){
    nodes[2*i].value = LCP[i-1];
    nodes[2*i+1].value = n-SA[i]+1; // length of string including 1 past end
    nodes[2*i].parent = 0;
    nodes[2*i+1].parent = 0;
  }
  nodes[0].value = 0;
  nodes[1].value = n-SA[0]+1;
  nodes[0].parent = nodes[1].parent = 0;
}

void run(node* nodes, long n){
  long m = 2*n;
  cartesianTree(nodes, 1, m-1);
  //nextTime("Time for building CT in parallel");

  // shortcut to roots of each cluster
  cilk_for(long i = 1; i < m; i++)
    nodes[i].parent = getRoot(nodes, i);
  //nextTime("Time for shortcuts");
}

};

#endif
