/* This code is part of the paper
 "Space-efficient Parallel Construction of Succinct Representations
 of Suffix Tree Topologies"
 by Uwe Baier, Timo Beller and Enno Ohlebusch (to be published in JEA 2016)
 
 Copyright (c) 2016 Uwe Baier
 
 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to deal
 in the Software without restriction, including without limitation the
 rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 sell copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:
 
 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.
 
 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 SOFTWARE.
*/
#include <algorithm>
#include <array>
#include <bitset>
#include <chrono>
#include <limits>
#include <stdexcept>
#include <stdint.h>
#include <string>
#include <iostream>
#include <vector>

#include <sdsl/bits.hpp>
#include <sdsl/int_vector.hpp>
#include <sdsl/io.hpp>
#include <sdsl/cst_sct3.hpp>
#include <sdsl/sorted_stack_support.hpp>
#include <sdsl/sorted_multi_stack_support.hpp>

#include <omp.h>

#include <cilk/cilk.h>
#include <cilk/cilk_api.h>

#include "../include/bps_dfs.hpp"
#include "../include/bps_dfs_par_opt.hpp"
#include "../include/bps_dfs_par_heur.hpp"
#include "../include/bps_sct.hpp"
#include "../include/bps_sct_par_opt.hpp"
#include "../include/bps_sct_par_heur.hpp"
#include "../include/shun.h"

using namespace std;
using namespace sdsl;
using namespace std::chrono;

typedef high_resolution_clock timer;

#ifndef NUM_RUNS
#define NUM_RUNS 1
#endif

void printUsage( const char *cmd ) {
	cerr << "USAGE: " << cmd << " [PROCESSORS] TESTCASEFILE" << endl;
	cerr << "- PROCESSORS: a list of the number of processors to be used." << endl;
	cerr << "              Note that processor count p=1 always is included." << endl;
	cerr << "- TESTCASEFILE: path to file used for benchmarking." << endl;
	cerr << "  Results are measured in milliseconds, and are average of " << NUM_RUNS << " runs." << endl;
}

typedef uint64_t value_type;
typedef uint64_t size_type;

int main(int argc, char **argv) {
	//benchmark setup
	const size_type a_bpsdfsseq = 0;
	const size_type a_bpsdfsseqbs = 1;
	const size_type a_bpsdfssdsl = 2;
	const size_type a_bpsdfsparopt = 3;
	const size_type a_bpsdfsparheur = 4;
	const size_type a_bpssctseq = 5;
	const size_type a_bpssctseqbs = 6;
	const size_type a_bpssctparopt = 7;
	const size_type a_bpssctparheur = 8;
	const size_type a_mwsctshun = 9;
	const size_type numalgorithms = 10;
	const size_type runs = NUM_RUNS;
	array<string,numalgorithms> alglabel;
	alglabel[a_bpsdfsseq] = "bpsdfsseq";
	alglabel[a_bpsdfsseqbs] = "bpsdfsseqbs";
	alglabel[a_bpsdfssdsl] = "bpsdfssdsl";
	alglabel[a_bpsdfsparopt] = "bpsdfsparopt";
	alglabel[a_bpsdfsparheur] = "bpsdfsparheur";
	alglabel[a_bpssctseq] = "bpssctseq";
	alglabel[a_bpssctseqbs] = "bpssctseqbs";
	alglabel[a_bpssctparopt] = "bpssctparopt";
	alglabel[a_bpssctparheur] = "bpssctparheur";
	alglabel[a_mwsctshun] = "mwsctshun";
	//timings from current processor-count run
	array<uint64_t,numalgorithms> timings;
	//timings from last processor-count run (for sequential algorithms)
	array<uint64_t,numalgorithms> ltimings;

	//process arguments
	if (argc < 2) {
		printUsage(argv[0]);
		return 1;
	}
	//fetch number of processors to be used
	vector<size_type> processors;	processors.reserve( argc - 1 );
	processors.push_back( (size_type)1 );
	for (int i = 1; i < argc - 1; i++) {
		try {
			//read number
			string arg{ argv[i] };
			size_t idx = 0;
			size_type p = (size_type)stoul( arg, &idx );
			if (idx != arg.length() || p == 0)	{
				throw invalid_argument("processor size must be greater zero");
			}
			//ensure p is not contained already
			if (find( processors.begin(), processors.end(), p) == processors.end()) {
				processors.push_back( p );
			}
		} catch (invalid_argument &e) {
			printUsage( argv[0] );
			cerr << "invalid value for processor count" << endl;
			return 1;
		}
	}

	//set up lcp array from testcase
	string tcfile( argv[argc-1] );
	string tcname = util::basename( tcfile );

	cache_config config(true, "@" ); //construct store files in memory
	size_type sigma = 0;
	size_type n = 0;
	//load text and collect information
	{
		const char *KEY_TEXT = key_text_trait<8>::KEY_TEXT;
		int_vector<8> text;

		//load text
		if (!load_vector_from_file(text, tcfile, 1)) {
			cerr << "Text could not be loaded" << endl;
			return 1;
		}
		if (contains_no_zero_symbol(text, tcfile)) {
			append_zero_symbol(text);
			store_to_cache(text,KEY_TEXT, config);
		}
		store_to_cache(text, KEY_TEXT, config);

		//collect text information
		n = text.size();
		if (n >= numeric_limits<value_type>::max()) {
			printUsage( argv[0] );
			cerr << "Filesize must be smaller than " 
			     << numeric_limits<value_type>::max() << " bytes" << endl;
		}
		bitset<256> C;
		for (size_type i = 0; i < n; i++) {
			if (!C[text[i]]) {
				++sigma;	C[text[i]] = true;
			}
		}
		--sigma; //ignore zero symbol at end
	}

	//construct suffix array if not cached yet
	if (!cache_file_exists(conf::KEY_SA, config)) {
		construct_sa<8>( config );
	} else {
		register_cache_file(conf::KEY_SA, config);
	}

	//construct lcp array
	if (!cache_file_exists(conf::KEY_LCP, config)) {
		construct_lcp_PHI<8>(config);
	} else {
		register_cache_file(conf::KEY_LCP, config);
	}

	//get in-memory-buffered sa and lcp array
	int_vector_buffer<> sa_buf(cache_file_name(conf::KEY_SA, config));
	int_vector_buffer<> lcp_buf(cache_file_name(conf::KEY_LCP, config));

	//prepare buffers for correct sequences if required
#ifdef CHECKCORRECTNESS
	bit_vector bps_check;
	bit_vector bpr_check;
	bit_vector b_check;

	size_type bps_check_origin; //to indicate from which algorithm the check originated
	size_type ebpr_check_origin;
#endif

	//output text information
	cout << "#filesize: " << n << endl;
	cout << "#sigma: " << sigma << endl;
	//output labels
	cout << "processors";
	for (size_type i = 0; i < timings.size(); i++) {
		cout << "\t" << alglabel[i];
	}
	cout << endl;

	for (auto p : processors) {
		for (size_type i = 0; i < timings.size(); i++) {
			ltimings[i] = timings[i];
			timings[i] = 0;
		}
		for (size_type i = 0; i < runs; i++) {

			//bps dfs sequential algorithm
			if (p == 1) {
				bit_vector bps( 4*n );
				vector<value_type> lcp( lcp_buf.begin(), lcp_buf.end() );

				auto start = timer::now();
				bps_dfs<decltype(lcp), value_type, size_type,
				         std::sorted_stack<value_type>>( lcp, bps );

				auto end = timer::now();
				timings[a_bpsdfsseq] += (uint64_t)duration_cast<milliseconds>( end - start ).count();

#ifdef CHECKCORRECTNESS
				bps_check = bps;
				bps_check_origin = a_bpsdfsseq;
#endif
			} else {
				timings[a_bpsdfsseq] += ltimings[a_bpsdfsseq];
			}

			//bps dfs sequential algorithm with bitstack
			if (p == 1) {
				bit_vector bps( 4*n );
				vector<value_type> lcp( lcp_buf.begin(), lcp_buf.end() );

				auto start = timer::now();
				bps_dfs<decltype(lcp), value_type, size_type,
				         sorted_stack_support>( lcp, bps );

				auto end = timer::now();
				timings[a_bpsdfsseqbs] += (uint64_t)duration_cast<milliseconds>( end - start ).count();

#ifdef CHECKCORRECTNESS
				if (bps != bps_check) {
					cerr << "BPS from algorithms " << alglabel[bps_check_origin]
					     << " and " << alglabel[a_bpsdfsseqbs] << " differ";
					return 1;
				}
#endif
			} else {
				timings[a_bpsdfsseqbs] += ltimings[a_bpsdfsseqbs];
			}

			//bps dfs sdsl algorithm
			if (p == 1) {
				bit_vector bps( 4*n );

				auto start = timer::now();
				sdsl::cst_sct3<> temp_cst(config, true);
				bps.resize(4*(temp_cst.bp.size()/2));
				util::set_to_value(bps, 0);
				uint64_t idx=0;
				for (auto it=temp_cst.begin(), end=temp_cst.end(); it!=end; ++it) {
					if (1 == it.visit())
						bps[idx] = 1;
					if (temp_cst.is_leaf(*it) and temp_cst.root()!= *it)
						++idx;
					++idx;
				}
				bps.resize(idx);
				auto end = timer::now();
				timings[a_bpsdfssdsl] += (uint64_t)duration_cast<milliseconds>( end - start ).count();

#ifdef CHECKCORRECTNESS
				if (bps != bps_check) {
					cerr << "BPS from algorithms " << alglabel[bps_check_origin]
					     << " and " << alglabel[a_bpsdfssdsl] << " differ";
					return 1;
				}
#endif
			} else {
				timings[a_bpsdfssdsl] += ltimings[a_bpsdfssdsl];
			}

			//bps dfs parallel optimal algorithm
			{
				bit_vector bps( 4*n );
				vector<value_type> lcp( lcp_buf.begin(), lcp_buf.end() );
			
				auto start = timer::now();
				bps_dfs_par_opt<value_type>( p, lcp, bps );
				auto end = timer::now();
				timings[a_bpsdfsparopt] += (uint64_t)duration_cast<milliseconds>( end - start ).count();

#ifdef CHECKCORRECTNESS
				if (bps != bps_check) {
					cerr << "BPS from algorithms " << alglabel[bps_check_origin]
					     << " and " << alglabel[a_bpsdfsparopt] << " differ";
					return 1;
				}
#endif
			}

			//bps dfs parallel heuristical algorithm
			{
				bit_vector bps( 4*n );
				vector<value_type> lcp( lcp_buf.begin(), lcp_buf.end() );

				auto start = timer::now();
				bps_dfs_par_heur<decltype(lcp), value_type, size_type,
				         decltype(bps), bit_vector::size_type>( p, lcp, bps );
				auto end = timer::now();
				timings[a_bpsdfsparheur] += (uint64_t)duration_cast<milliseconds>( end - start ).count();

#ifdef CHECKCORRECTNESS
				if (bps != bps_check) {
					cerr << "BPS from algorithms " << alglabel[bps_check_origin]
					     << " and " << alglabel[a_bpsdfsparheur] << " differ";
					return 1;
				}
#endif
				
			}

			//bps sct sequential algorithm
			if (p == 1) {
				bit_vector bpr( 2*n );
				bit_vector b( n );
				vector<value_type> lcp( lcp_buf.begin(), lcp_buf.end() );

				auto start = timer::now();
				bps_sct<decltype(lcp), value_type, size_type,
				        std::sorted_multi_stack<value_type>>( lcp, bpr, b );
				auto end = timer::now();
				timings[a_bpssctseq] += (uint64_t)duration_cast<milliseconds>( end - start ).count();

#ifdef CHECKCORRECTNESS
				bpr_check = bpr;
				b_check = b;
				ebpr_check_origin = a_bpssctseq;
#endif
			} else {
				timings[a_bpssctseq] += ltimings[a_bpssctseq];
			}

			//bps sct sequential algorithm with bitstack
			if (p == 1) {
				bit_vector bpr( 2*n );
				bit_vector b( n );
				vector<value_type> lcp( lcp_buf.begin(), lcp_buf.end() );

				auto start = timer::now();
				bps_sct<decltype(lcp), value_type, size_type,
				        sorted_multi_stack_support>( lcp, bpr, b );
				auto end = timer::now();
				timings[a_bpssctseqbs] += (uint64_t)duration_cast<milliseconds>( end - start ).count();

#ifdef CHECKCORRECTNESS
				if (bpr != bpr_check || b != b_check) {
					cerr << "eBPR from algorithms " << alglabel[ebpr_check_origin]
					     << " and " << alglabel[a_bpssctseqbs] << " differ";
					return 1;
				}
#endif
			} else {
				timings[a_bpssctseqbs] += ltimings[a_bpssctseqbs];
			}

			//bps sct parallel optimal algorithm
			{
				bit_vector bpr( 2*n );
				bit_vector b( n );
				vector<value_type> lcp( lcp_buf.begin(), lcp_buf.end() );

				auto start = timer::now();
				bps_sct_par_opt<value_type>( p, lcp, bpr, b );
				auto end = timer::now();
				timings[a_bpssctparopt] += (uint64_t)duration_cast<milliseconds>( end - start ).count();

#ifdef CHECKCORRECTNESS
				if (bpr != bpr_check || b != b_check) {
					cerr << "eBPR from algorithms " << alglabel[ebpr_check_origin]
					     << " and " << alglabel[a_bpssctparopt] << " differ";
					return 1;
				}
#endif
			}

			//bps sct parallel heuristical algorithm
			{
				bit_vector bpr( 2*n );
				bit_vector b( n );
				vector<value_type> lcp( lcp_buf.begin(), lcp_buf.end() );

				auto start = timer::now();
				bps_sct_par_heur<decltype(lcp), value_type, size_type,
				         decltype(bpr), bit_vector::size_type>( p, lcp, bpr, b );
				auto end = timer::now();
				timings[a_bpssctparheur] += (uint64_t)duration_cast<milliseconds>( end - start ).count();

#ifdef CHECKCORRECTNESS
				if (bpr != bpr_check || b != b_check) {
					cerr << "eBPR from algorithms " << alglabel[ebpr_check_origin]
					     << " and " << alglabel[a_bpssctparheur] << " differ" << endl;
					return 1;
				}
#endif
			}

			//parallel multiway super-cartesian tree by shun
			{
				//do some cilk setup
				auto p_str = to_string( p );
				__cilkrts_end_cilk();
				__cilkrts_set_param("nworkers",p_str.c_str());

				//and create necessary structures
				vector<value_type> sa( sa_buf.begin(), sa_buf.end() );
				vector<value_type> lcp( lcp_buf.begin(), lcp_buf.end() );
				vector<shun::node> nodes( 2*lcp.size() );
				
				//run algorithm
				auto start = timer::now();
				shun::prepare( sa.data(), lcp.data(), (long)lcp.size(), nodes.data() );
				shun::run( nodes.data(), (long)lcp.size() );
				auto end = timer::now();
				timings[a_mwsctshun] += (uint64_t)duration_cast<milliseconds>( end - start ).count();
			}
		}
		//output average results
		cout << p;
		for (size_t i = 0; i < timings.size(); i++) {
			timings[i] /= (uint64_t)runs;
			cout << "\t" << (timings[i]);
		}
		cout << endl;
	}
	return 0;
}
